local myNAME = "merGuildResearchNotes"
local myMODULES = merGuildResearchNotes.MODULES

local Main = myMODULES.Main
local SavedVars = myMODULES.SavedVars
local Settings = myMODULES.Settings
local Utils = myMODULES.Utils
local df = Utils.df

local MAX_TOOLTIP_LINES = 10


local function getDisplayNameForTooltip(memberName)
    local displayName = ZO_SECOND_CONTRAST_TEXT:Colorize(memberName)
    local charInfo = ""
    local guildId = SavedVars.get("guildMembers", memberName, "guildId")
    local memberIndex = SavedVars.get("guildMembers", memberName, "memberIndex")

    if guildId and memberIndex then
        local foundName, _, _, _, offline = GetGuildMemberInfo(guildId, memberIndex)
        -- safe-guard against saved (guildId, memberIndex) mismatch
        if foundName == memberName then
            local hasChar, charName = GetGuildMemberCharacterInfo(guildId, memberIndex)
            if hasChar then
                charInfo = zo_strformat(" <<1>>", charName)
            end
            charInfo = charInfo .. " (" .. GetGuildName(guildId) .. ")"
            -- dim offline members' character (guild) info
            if offline and offline > 0 then
                charInfo = ZO_DISABLED_TEXT:Colorize(charInfo)
            end
        end
    end

    return displayName .. charInfo
end


local function isResearchRowTraitKnown(row)
    local _, _, known = GetSmithingResearchLineTraitInfo(row.craftingType,
                                                         row.researchLineIndex,
                                                         row.traitIndex)
    return known
end


local function sortByKey(tab, key, desc)
    if desc then
        table.sort(tab, function(a, b) return b[key] < a[key] end)
    else
        table.sort(tab, function(a, b) return a[key] < b[key] end)
    end
end


local function postSmithingResearchRowMouseEnter(self, row)
    if not row or isResearchRowTraitKnown(row) then
        return
    end

    local guildLine = SavedVars.get("guildResearch", row.craftingType, row.researchLineIndex)
    local knownBy = {}

    if type(guildLine) == "table" then
        for name, knownTraitsBits in pairs(guildLine) do
            if Utils.testbit(knownTraitsBits, row.traitIndex - 1) then
                local time = SavedVars.get("guildMembers", name, "encResearchTime")
                local entry = {name = name, time = tonumber(time) or 0}
                table.insert(knownBy, entry)
            end
        end
    end

    if #knownBy == 0 then
        return
    end

    ZO_Tooltip_AddDivider(InformationTooltip)
    InformationTooltip:AddLine("Known by", "", ZO_TOOLTIP_DEFAULT_COLOR:UnpackRGB())

    sortByKey(knownBy, "time", true) -- ORDER BY time DESC

    -- cut off entries older than the MAX_TOOLTIP_LINES'th time
    while #knownBy > MAX_TOOLTIP_LINES do
        if knownBy[#knownBy].time == knownBy[MAX_TOOLTIP_LINES].time then
            break
        end
        table.remove(knownBy)
    end

    sortByKey(knownBy, "name", false) -- ORDER BY name ASC

    for i, entry in ipairs(knownBy) do
        local displayName = getDisplayNameForTooltip(entry.name)
        InformationTooltip:AddLine(displayName, "")
        if i >= MAX_TOOLTIP_LINES then
            break
        end
    end
end


local function postSmithingResearchShown(self)
    if SavedVars.DEBUG then
        Utils.timeIt(Main, "refreshGuildMemberNotes", 15)
    else
        Main.refreshGuildMemberNotes(300)
    end
end


local function onResearchCompleted(eventCode, craftingType, lineIndex, traitIndex)
    local charName = GetUnitName("player")
    local charLines = SavedVars.sub("characters", charName,
                                    "craftingSkills", craftingType,
                                    "researchLines")
    local aggLine = SavedVars.sub("aggregateResearch", craftingType, lineIndex)

    -- replace saved research line info with newly scanned one
    local charLine = Main.scanCraftingResearchLine(craftingType, lineIndex)
    charLines[lineIndex] = assert(charLine)

    if SavedVars.DEBUG then
        Utils.timeIt(Main, "updateAggregateResearchLine", aggLine, charLine)
        Utils.timeIt(Main, "updatePlayerGRNs")
    else
        Main.updateAggregateResearchLine(aggLine, charLine)
        Main.updatePlayerGRNs()
    end
end


local function onPlayerActivated()
    if SavedVars.DEBUG then
        Utils.timeIt(Main, "updateAggregateResearch")
        Utils.timeIt(Main, "updatePlayerGRNs")
    else
        Main.updateAggregateResearch()
        Main.updatePlayerGRNs()
    end
    -- full update is only needed once
    EVENT_MANAGER:UnregisterForEvent(myNAME, EVENT_PLAYER_ACTIVATED)
end


local function onAddOnLoaded(event, addOnName)
    if addOnName ~= myNAME then return end
    EVENT_MANAGER:UnregisterForEvent(myNAME, EVENT_ADD_ON_LOADED)

    SavedVars.init(myNAME .. "_SavedVariables")
    SavedVars.upgrade()

    EVENT_MANAGER:RegisterForEvent(myNAME, EVENT_PLAYER_ACTIVATED,
                                   Utils.pwrap(onPlayerActivated))
    EVENT_MANAGER:RegisterForEvent(myNAME, EVENT_SMITHING_TRAIT_RESEARCH_COMPLETED,
                                   Utils.pwrap(onResearchCompleted))

    Utils.postHook(SMITHING.researchPanel, "OnResearchRowMouseEnter",
                   Utils.pwrap(postSmithingResearchRowMouseEnter))
    Utils.postHookHandler(SMITHING.researchPanel.control, "OnEffectivelyShown",
                          Utils.pwrap(postSmithingResearchShown))

    Settings.init()
end


EVENT_MANAGER:RegisterForEvent(myNAME, EVENT_ADD_ON_LOADED, onAddOnLoaded)

