local myMODULES = merGuildResearchNotes.MODULES
local Main = myMODULES.Main

local SavedVars = myMODULES.SavedVars
local Utils = myMODULES.Utils
local iterExisting = Utils.iterExisting
local max = math.max

local GRAY = "|cc5c5c5"
local GREEN = "|c1afa1a"
local LGREEN = "|c7afa7a"
local LPURPLE = "|cda7afa"
local LYELLOW = "|cfafa7a"
local ORANGE = "|cfaaf00"
local RED = "|cfa1a1a"
local WHITE = "|cffffff"

--[[
    BASE64 ENCODING CHARACTER SET

    Swapped some characters for purely cosmetic reasons:
    '-' represents no traits known on any character
    '+' represents all traits known on any single character
--]]
local BASE64_CHARACTERS = "-123456789ABCDEF+HIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0G"
local CURRENT_ENCODING_VERSION = "a"

--[[
    If the order of research lines (in any crafting skill) or the order of traits
    (in any research line) ever changes, OLDEST_ACCEPTED_TIMESTAMP needs to be set
    to some time after the patch that changed it goes live (along with API version
    bump). This is necessary because the current encoding relies on these orders.

    For example if I have Dagger:Precise=known and Sword:Precise=unknown, and ZOS
    swaps the indices of Dagger and Sword, then until I update my GRN, people with
    out-of-date add-on will see me having Sword:Precise=known. But people with
    updated add-on will ignore my GRN because the timestamp it contains will be
    older than their OLDEST_ACCEPTED_TIMESTAMP.
--]]
local OLDEST_ACCEPTED_TIMESTAMP = 1405026899

local CRAFTING_TYPES = {CRAFTING_TYPE_BLACKSMITHING,
                        CRAFTING_TYPE_CLOTHIER,
                        CRAFTING_TYPE_WOODWORKING}

local CRAFTING_TYPE_KEYS = {[CRAFTING_TYPE_BLACKSMITHING] = "B",
                            [CRAFTING_TYPE_CLOTHIER] = "C",
                            [CRAFTING_TYPE_WOODWORKING] = "W"}

local NUM_TRAITS = 8


local function knownBitsToString(craftingType, lineIndex, knownBits)
    local str = ""
    for traitIndex, traitType, traitName in
        Utils.iterSmithingResearchLineTraits(craftingType, lineIndex) do
        if Utils.testbit(knownBits, traitIndex - 1) then
            str = str .. " " .. traitType .. ":" .. traitName
        end
    end
    return str:sub(2) -- skip the first space
end


function Main.decodeResearchChunk(encChunk)
    local craftingKey = encChunk:sub(1, 1)
    local craftingType = nil
    for ct, ck in pairs(CRAFTING_TYPE_KEYS) do
        if craftingKey == ck then
            craftingType = ct
            break
        end
    end
    if not craftingType then
        error("unknown encoded data prefix '" .. craftingKey .. "'")
    end

    local encPos = 2
    local res = {}

    local function decError(offset)
        local msg = "invalid encoded data %q[%d]"
        error(msg:format(encData, encPos + offset))
    end

    local function decNextChar()
        local chr = encChunk:sub(encPos, encPos)
        local idx = BASE64_CHARACTERS:find(chr)
        if (not idx) or (idx < 1) or (idx > 64) then
            decError(0)
        end
        encPos = encPos + 1
        return idx - 1
    end

    local numLines = GetNumSmithingResearchLines(craftingType)
    for lineIndex = 1, numLines do
        local c0 = decNextChar()
        local knownBits = 0
        local subCharKnown = 0
        if c0 >= 0x20 then
            local k1, k0 = Utils.splitbits(c0, 5, 2, 3)
            local k2 = decNextChar()
            knownBits = k1 * 64 + k2
            subCharKnown = k0
        elseif c0 >= 0x10 then
            -- all eight traits known
            knownBits = 255
            subCharKnown = c0 % 8
        elseif c0 == 0x00 then
            -- no traits known
            knownBits = 0
            subCharKnown = 0
        elseif c0 <= 0x08 then
            -- only one trait known
            knownBits = Utils.setbit(0, c0 - 1)
            subCharKnown = 0
        else
            decError(-1)
        end

        local numKnown = Utils.countsetbits(knownBits)
        local lineInfo = {
            knownBits = knownBits,
            maxCharKnown = numKnown - subCharKnown,
        }

        if SavedVars.DEBUG then
            lineInfo.knownList = knownBitsToString(craftingType, lineIndex, knownBits)
            lineInfo.numKnown = numKnown
            lineInfo.subCharKnown = subCharKnown
            lineInfo.lineName = GetSmithingResearchLineInfo(craftingType, lineIndex)
        end

        res[lineIndex] = lineInfo
    end

    return craftingType, res
end


function Main.encodeAggregateResearch()
    local aggResearch = SavedVars.sub("aggregateResearch")
    local encData = ""

    for _, craftingType in ipairs(CRAFTING_TYPES) do
        local encChunk = Main.encodeAggregateResearchCraft(craftingType, aggResearch)
        encData = encData .. ":" .. encChunk
    end

    return "#!GRN:" .. CURRENT_ENCODING_VERSION .. GetTimeStamp() .. encData .. "!"
end


function Main.encodeAggregateResearchCraft(craftingType, aggResearch)
    local encChunk = assert(CRAFTING_TYPE_KEYS[craftingType])
    local function encNextChar(val)
        local idx = 1 + val
        encChunk = encChunk .. BASE64_CHARACTERS:sub(idx, idx)
    end

    local numLines = GetNumSmithingResearchLines(craftingType)
    for lineIndex = 1, numLines do
        local aggLine = Utils.subtable(aggResearch, craftingType, lineIndex)
        local tailPos = #encChunk + 1
        aggLine.maxCharKnown = aggLine.maxCharKnown or 0
        aggLine.knownBits = aggLine.knownBits or 0
        aggLine.numKnown = Utils.countsetbits(aggLine.knownBits)
        if aggLine.numKnown == 0 then
            encNextChar(0x00)
        elseif aggLine.numKnown == 1 then
            encNextChar(aggLine.lastKnown)
        elseif aggLine.numKnown == 8 then
            local k0 = aggLine.numKnown - aggLine.maxCharKnown
            aggLine.subCharKnown = k0
            encNextChar(0x10 + (k0 % 8))
        else
            local k1, k2 = Utils.splitbits(aggLine.knownBits, 8, 2, 6)
            local k0 = aggLine.numKnown - aggLine.maxCharKnown
            aggLine.subCharKnown = k0
            aggLine.knownBits76 = k1
            aggLine.knownBits50 = k2
            encNextChar(0x20 + (k0 % 8) + (k1 * 8))
            encNextChar(k2)
        end
        aggLine.encChunk = encChunk:sub(tailPos)
    end
    return encChunk
end


function Main.getAggregateGRN()
    local GRN = SavedVars.aggregateGRN
    if type(GRN) ~= "string" then
        GRN = Main.encodeAggregateResearch()
        SavedVars.aggregateGRN = GRN
    end
    return GRN
end


function Main.parseMemberNote(note)
    if type(note) == "string" then
        local ms, me, ver, time, data = note:find("#!GRN:([a-z]+)(%d+)([^#!]*)!")
        return ms, me, ver, tonumber(time), data
    end
end


function Main.pruneGuildResearch()
    local guildMembers = SavedVars.sub("guildMembers")

    for _, craftingType in ipairs(CRAFTING_TYPES) do
        local guildLines = SavedVars.get("guildResearch", craftingType)
        local numLines = GetNumSmithingResearchLines(craftingType)

        for lineIndex, guildLine in iterExisting(guildLines, 1, numLines) do
            for name in pairs(guildLine) do
                if not guildMembers[name] then
                    guildLine[name] = nil
                end
            end
        end
    end
end


function Main.refreshGuildMemberNotes(staleSeconds)
    local currentTime = GetTimeStamp()
    local staleThreshold = currentTime - (tonumber(staleSeconds) or 300)

    if staleThreshold < (SavedVars.lastRefresh or 0) then
        return
    end

    SavedVars.lastRefresh = currentTime

    for guildIndex = 1, MAX_GUILDS do
        local guildId = GetGuildId(guildIndex)
        local guildName = GetGuildName(guildId)
        if SavedVars.get("guildSettings", guildName, "enabled") then
            Main.scanGuildMemberNotes(guildId, currentTime)
        end
    end

    Main.updateGuildResearch()
    Main.pruneGuildResearch()
end


function Main.scanCraftingResearch(craftingType)
    local name = GetSkillLineInfo(GetCraftingSkillLineIndices(craftingType))
    local numLines = GetNumSmithingResearchLines(craftingType)
    local res = nil

    for lineIndex = 1, numLines do
        local lineInfo = Main.scanCraftingResearchLine(craftingType, lineIndex)
        if lineInfo ~= nil then
            if res == nil then
                res = {}
                res.craftingSkillName = name
                res.researchLines = {}
            end
            res.researchLines[lineIndex] = lineInfo
        end
    end
    return res
end


function Main.scanCraftingResearchLine(craftingType, lineIndex)
    local lineName, _, numTraits = GetSmithingResearchLineInfo(craftingType, lineIndex)
    local knownBits = 0
    local res = nil

    for traitIndex = 1, numTraits do
        local _, _, known = GetSmithingResearchLineTraitInfo(craftingType, lineIndex, traitIndex)
        if known then
            knownBits = Utils.setbit(knownBits, traitIndex - 1)
        end
    end

    -- if there are some known traits, return table, otherwise return nil
    if knownBits > 0 then
        res = {}
        res.researchLineName = lineName
        res.knownBits = knownBits
        if SavedVars.DEBUG then
            res.knownList = knownBitsToString(craftingType, lineIndex, knownBits)
        end
    end
    return res
end


function Main.scanGuildMemberNotes(guildId, currentTime)
    local guildMembers = SavedVars.sub("guildMembers")
    local currentTime = currentTime or GetTimeStamp()

    for memberIndex = 1, GetNumGuildMembers(guildId) do
        local name, note = GetGuildMemberInfo(guildId, memberIndex)
        local _, _, ver, time, data = Main.parseMemberNote(note)
        if ver ~= CURRENT_ENCODING_VERSION then
            -- if I need to change the encoding for whatever reason,
            -- I will certaintly not be willing to drag old decoding
            -- functions along, so to KISS, ignore members using old
            -- add-on versions
        elseif (time or 0) >= OLDEST_ACCEPTED_TIMESTAMP then
            local member = Utils.subtable(guildMembers, name)
            if (member.lastRefresh or 0) < currentTime then
                member.lastRefresh = currentTime
                member.guildId = guildId
                member.memberIndex = memberIndex
            end
            if (member.encResearchTime or 0) < time then
                member.encResearchTime = time
                member.encResearchData = data:gsub("%s+", "")
            end
        end
    end
end


function Main.setPlayerGRN(guildId, newGRN)
    if DoesPlayerHaveGuildPermission(guildId, GUILD_PERMISSION_NOTE_EDIT) then
        local memberIndex = GetPlayerGuildMemberIndex(guildId)
        local _, oldNote = GetGuildMemberInfo(guildId, memberIndex)
        local ms, me = Main.parseMemberNote(oldNote)
        -- insert line breaks and spaces for readability
        local newNote = newGRN:gsub(":([A-Z])", ":\n  %1")

        if ms and me then
            -- oldNote already contains GRN, which must be replaced,
            -- while keeping the text before and after it untouched
            newNote = oldNote:sub(1, ms - 1) .. newNote .. oldNote:sub(me + 1)
        elseif oldNote and oldNote ~= "" then
            newNote = oldNote .. "\n" .. newNote
        end
        SetGuildMemberNote(guildId, memberIndex, newNote)
    end
end


function Main.updateAggregateResearch()
    local charResearch = Main.updateCharacterResearch()
    local aggResearch = SavedVars.sub("aggregateResearch")

    for _, craftingType in ipairs(CRAFTING_TYPES) do
        local numLines = GetNumSmithingResearchLines(craftingType)
        local charLines = Utils.gettable(charResearch, "craftingSkills",
                                         craftingType, "researchLines")
        for lineIndex, charLine in iterExisting(charLines, 1, numLines) do
            local aggLine = Utils.subtable(aggResearch, craftingType, lineIndex)
            Main.updateAggregateResearchLine(aggLine, charLine)
        end
    end
end


function Main.updateAggregateResearchLine(aggResearchLine, charResearchLine)
    local aggKnownBits = aggResearchLine.knownBits or 0
    local charKnownBits = charResearchLine.knownBits or 0
    local numCharKnown = 0

    for traitIndex = 1, NUM_TRAITS do
        if Utils.testbit(charKnownBits, traitIndex - 1) then
            aggKnownBits = Utils.setbit(aggKnownBits, traitIndex - 1)
            numCharKnown = numCharKnown + 1
            aggResearchLine.lastKnown = traitIndex
        end
    end
    aggResearchLine.knownBits = aggKnownBits
    aggResearchLine.maxCharKnown = max(numCharKnown, aggResearchLine.maxCharKnown or 0)
    SavedVars.aggregateGRN = nil -- invalidate cached GRN
end


function Main.updateCharacterResearch()
    local charName = GetUnitName("player")
    local charResearch = SavedVars.sub("characters", charName)
    charResearch.craftingSkills = {}

    for _, craftingType in ipairs(CRAFTING_TYPES) do
        local craftingSkill = Main.scanCraftingResearch(craftingType)
        charResearch.craftingSkills[craftingType] = craftingSkill
    end
    return charResearch
end


function Main.updateGuildMemberResearch(name, member)
    local encData = member.encResearchData
    local decData = {}

    if type(encData) ~= "string" then
        return
    end

    for encChunk in encData:gmatch(":%s*([A-Z][-+0-9A-Za-z]*)") do
        rawset(decData, Main.decodeResearchChunk(encChunk))
    end

    if SavedVars.DEBUG then
        member.decResearchData = decData
    end

    for _, craftingType in ipairs(CRAFTING_TYPES) do
        local decLines = Utils.gettable(decData, craftingType)
        local numLines = GetNumSmithingResearchLines(craftingType)
        for lineIndex, decLine in iterExisting(decLines, 1, numLines) do
            local guildLine = SavedVars.sub("guildResearch", craftingType, lineIndex)
            if (decLine.knownBits or 0) > 0 then
                guildLine[name] = decLine.knownBits
            else
                guildLine[name] = nil
            end
        end
    end
end


function Main.updateGuildResearch()
    local guildMembers = SavedVars.sub("guildMembers")
    local lastRefresh = SavedVars.lastRefresh or 0

    for name, member in pairs(guildMembers) do
        if (member.lastRefresh or 0) < lastRefresh then
            guildMembers[name] = nil
        else
            Main.updateGuildMemberResearch(name, member)
        end
    end
end


function Main.updatePlayerGRN(guildId)
    local guildName = GetGuildName(guildId)
    if SavedVars.get("guildSettings", guildName, "enabled") then
        Main.setPlayerGRN(guildId, Main.getAggregateGRN())
    else
        Main.setPlayerGRN(guildId, "")
    end
end


function Main.updatePlayerGRNs()
    for guildIndex = 1, MAX_GUILDS do
        local guildId = GetGuildId(guildIndex)
        Main.updatePlayerGRN(guildId)
    end
end

